from datetime import datetime
from typing import Any, Mapping, Optional, Union, Dict, List, Iterator
from pymongo import InsertOne, MongoClient, ReplaceOne, UpdateOne  # type: ignore

TIMESTAMPS_COLLECTION = "processed_tables_timestamps"
PROCESSED_TABLES_TABLE_NAME_FIELD = "source"
TIMESTAMPS_FIELD_NAME = "timestamp"


class MongoDBDAO:
    def __init__(self, connection_string: str, database_name: str) -> None:
        self.connection_string = connection_string
        self.database_name = database_name
        self.client = None
        self.database = None
        self.collection = None
        self.operations: List[Any] = []

    def connect(self) -> None:
        """
        Create MongoDBClient and connect to database.
        """
        self.client = MongoClient(self.connection_string)
        self.database = self.client[self.database_name]  # type: ignore

    def disconnect(self) -> None:
        """
        Dereference database and collection, close client and dereference it.
        """
        self.database = None
        self.client.close()  # type: ignore
        self.client = None

    def get_one(
        self, collection_name: str, filter_query: Dict[str, Union[str, int, float]]
    ) -> Optional[Dict[str, Union[str, int, float]]]:
        """
        Get document corresponding to filter_query.
        """
        return self.database[collection_name].find_one(filter_query)  # type: ignore

    def get_many(
        self, collection_name: str, filter_query: Dict[str, Union[str, int, float]]
    ) -> Iterator[Dict[str, Union[str, int, float]]]:
        """
        Get document corresponding to filter_query.
        """
        return (dict(item) for item in self.database[collection_name].find(filter_query))  # type: ignore

    def update_one(
        self,
        collection_name: str,
        filter_query: Dict[str, Union[str, int, float]],
        updates: Dict[str, Union[str, int, float, datetime]],
    ) -> None:
        """
        Set fields in filtered document to
        values supplied in updates dictionary.
        If the document does not exist, create it.
        """
        self.database[collection_name].update_one(  # type: ignore
            filter_query, {"$set": updates}, upsert=True
        )

    def insert_many(self, collection_name: str, data: List[Dict[Any, Any]]) -> None:
        """
        Insert data to MongoDB.
        """
        self.database[collection_name].insert_many(data)  # type: ignore

    def delete_many(
        self, collection_name: str, filter_query: Dict[str, Union[str, int, float]]
    ) -> None:
        self.database[collection_name].delete_many(filter_query)  # type: ignore

    def delete_all(self, collection_name: str) -> None:
        self.delete_many(collection_name, {})

    def update_timestamp(self, table_name: str, timestamp: str) -> None:
        self.update_one(
            TIMESTAMPS_COLLECTION,
            {PROCESSED_TABLES_TABLE_NAME_FIELD: table_name},
            {TIMESTAMPS_FIELD_NAME: timestamp},
        )

    def last_timestamp(self, table_name: str) -> Union[str, None]:
        """
        Returns timestamp from last successful insertion.
        """
        document = self.get_one(
            TIMESTAMPS_COLLECTION, {PROCESSED_TABLES_TABLE_NAME_FIELD: table_name}
        )
        if document is None:
            return None  # type: ignore

        return document[TIMESTAMPS_FIELD_NAME]  # type: ignore

    def add_replace_operation(
        self, filter: Mapping[str, Any], replacement: Mapping[str, Any]
    ) -> None:
        self.operations.append(
            ReplaceOne(filter=filter, replacement=replacement, upsert=True)
        )

    def add_insert_operation(self, document) -> None:
        self.operations.append(InsertOne(document))

    def add_update_operation(
        self, filter: Mapping[str, Any], update: Mapping[str, Any], upsert: bool = False
    ) -> None:
        self.operations.append(UpdateOne(filter, update, upsert))

    def execute_operations(self, collection_name: str) -> None:
        if self.operations:
            self.database[collection_name].bulk_write(self.operations, ordered=False)  # type: ignore
        self.operations = []



from functools import wraps
from typing import Optional, Iterable, Union, Callable, Tuple, Dict, Union, List

import orjson
import redis
from redis import Sentinel
from redis.sentinel import MasterNotFoundError

NodeAddress = Tuple[str, int]
RecordValuesType = Dict[str, Union[str, int, float]]


class RetryException(Exception):
    """
    Raised when retry limit is exceeded.
    """

    def __init__(self, maximum_attempts: int) -> None:
        super().__init__(f"Maximum retry attempts of {maximum_attempts} exhausted.")


def retry_decorator(function) -> Callable:
    """
    Retry calling the wrapped method up to retry_attempts
    if any of handled_exceptions occurs. Raise
    RetryException if all retries fail.
    """

    @wraps(function)
    def wrapper(*args, **kwargs):
        cls = args[0]
        for attempt_number in range(cls.retry_attempts + 1):
            try:
                return function(*args, **kwargs)
            except cls.handled_exceptions as exc:
                if attempt_number < cls.retry_attempts:
                    cls.connect()
                else:
                    raise RetryException(cls.retry_attempts) from exc

    return wrapper


class RedisDAO:
    handled_exceptions = (MasterNotFoundError,)
    retry_attempts = 5

    def __init__(
        self,
        node_addresses: List[NodeAddress],
        password: Optional[str] = None,
        pseudo_schema_name: Optional[str] = None,
        master_name: str = "mymaster",
        use_sentinel: bool = True,
    ) -> None:
        self.node_addresses = node_addresses
        self.password = password
        self.master_name = master_name
        self.pseudo_schema_prefix = (
            f"{pseudo_schema_name}:" if pseudo_schema_name is not None else ""
        )
        self.use_sentinel = use_sentinel
        self.sentinels = None
        self.master = None

    def connect_sentinel(self) -> None:
        """
        Connect to Redis Sentinel cluster.
        """
        self.sentinels = Sentinel(
            self.node_addresses,
            socket_timeout=0.5,
            sentinel_kwargs={"password": self.password},
        )  # type: ignore
        host, port = self.sentinels.discover_master(self.master_name)  # type: ignore
        self.master = redis.StrictRedis(host=host, port=port, password=self.password)  # type: ignore

    def connect_single_node(self) -> None:
        """
        Connect to single node Redis instance.
        """
        self.master = redis.Redis(*self.node_addresses[0], password=self.password)  # type: ignore

    def connect(self) -> None:
        """
        Connect to a specified Redis instance.
        """
        if self.use_sentinel:
            self.connect_sentinel()
        else:
            self.connect_single_node()

    def disconnect(self) -> None:
        """
        De-reference Redis connection-related objects.
        """
        self.master = None
        self.sentinels = None

    def add_schema_to_key(self, key: str) -> str:
        """
        Prepend bare key with pseudo-schema.
        """
        return f"{self.pseudo_schema_prefix}{key}"

    @retry_decorator
    def get_one(self, key: str) -> Optional[str]:
        """
        Get value for key.
        """
        return self.master.get(self.add_schema_to_key(key))  # type: ignore

    @retry_decorator
    def get_many(self, keys: Iterable[str]) -> Dict[str, Optional[str]]:
        """
        Get key-value dicts for specified keys.,
        """
        return dict(
            zip(keys, self.master.mget(*(self.add_schema_to_key(key) for key in keys)))  # type: ignore
        )

    @retry_decorator
    def get_many_dicts(
        self, keys: Iterable[str]
    ) -> Dict[str, Optional[Dict[str, str]]]:
        """
        Get dict of key-dict for specified keys.
        """
        keys = list(keys)
        values = self.master.mget([self.add_schema_to_key(key) for key in keys])  # type: ignore
        return {
            key: (orjson.loads(value) if value is not None else None)
            for key, value in zip(keys, values)
        }

    @retry_decorator
    def update_one(self, key: str, value: str) -> bool:
        """
        Set value for key.
        """
        return self.master.set(self.add_schema_to_key(key), value)  # type: ignore

    @retry_decorator
    def update_many_if_unchanged(
        self,
        keys_values: Iterable[Tuple[str, RecordValuesType, RecordValuesType]],
        expiration_seconds: float,
    ) -> Dict[str, bool]:
        """
        Update each key in keys_values with new_value, if its current value
        matches expected_value.
        Return dict with keys and boolean update results (True if the key was updated).
        Operations for each key are atomic.
        """
        keys_values = list(keys_values)
        script = f"""
        local current_value = redis.call('get', KEYS[1])
        if current_value == ARGV[1] or current_value == false then
            redis.call('set', KEYS[1], ARGV[2])
            redis.call('expire', KEYS[1], {expiration_seconds}) 
            return 1
        end
        return 0
        """
        pipe = self.master.pipeline(transaction=False)  # type: ignore
        for key, expected_value, new_value in keys_values:
            pipe.eval(
                script,
                1,
                self.add_schema_to_key(key),
                orjson.dumps(expected_value, option=orjson.OPT_SORT_KEYS),
                orjson.dumps(new_value, option=orjson.OPT_SORT_KEYS),
            )
        result = pipe.execute()
        return {
            item[0]: bool(item_result) for item, item_result in zip(keys_values, result)
        }



import contextlib
import shutil
import time

from typing import Union, Generator, Optional


class Heartbeat:
    def __init__(self, filepath: str) -> None:
        self.filepath = filepath
        self.staging_filepath = f"{filepath}.staging"

    @staticmethod
    def _get_current_time() -> float:
        return time.time()

    def read(self) -> float:
        """
        Get time value from the heartbeat file.
        """
        with open(self.filepath, "r") as infile:
            last_time = infile.read()
        return float(last_time)

    def write_to_staging_file(self, heartbeat_content: str) -> None:
        """
        Write current time to the staging file.
        """
        with open(self.staging_filepath, "w", encoding="utf-8") as outfile:
            outfile.write(heartbeat_content)

    def write(self, heartbeat_content: str) -> None:
        """
        Write current time to the heartbeat file.
        First, write to a staging file and replace it
        when written. This is to avoid reading from empty file
        that is just being written into.
        """
        self.write_to_staging_file(heartbeat_content)
        shutil.move(self.staging_filepath, self.filepath)

    def update(self, heartbeat_content: Optional[str] = None) -> None:
        """
        Update heartbeat file to given content or current time.
        """
        heartbeat_content = (
            heartbeat_content
            if heartbeat_content is not None
            else str(self._get_current_time())
        )
        self.write(heartbeat_content)

    def is_alive(self, alive_threshold_ms: float) -> bool:
        """
        Check if timestamp in heartbeat file is newer than alive_threshold_ms.
        """
        last_time = self.read()
        return (self._get_current_time() - last_time) * 1000 < alive_threshold_ms

    @contextlib.contextmanager
    def long_operation(
        self, long_operation_duration_seconds: Union[int, float]
    ) -> Generator:
        """
        On entry, updates heartbeat to time which is
        long_operation_duration_seconds from now.
        On exit, updates heartbeat to current timestamp.
        """
        try:
            self.update(str(self._get_current_time() + long_operation_duration_seconds))
            yield
        finally:
            self.update()



from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="mallgroup_python_lib",
    version="0.0.10",
    description="Official Mallgroup Library",
    author="Data Engineering Mallgroup",
    author_email="dataengineering@mallgroup.com",
    packages=["mallgroup_python_lib"],
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "License :: OSI Approved :: MIT License",
    ],
    long_description=long_description,
    long_description_content_type="text/markdown",
    install_requires=[],
    extras_require={"dev": ["pytest"]},
)
